package jp.veltec.pdf;
import java.io.OutputStream;
import java.io.IOException;
import java.io.File;
import java.util.HashMap;
import java.net.MalformedURLException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.NodeList;

import com.itextpdf.text.Document;
import com.itextpdf.text.DocumentException;
import com.itextpdf.text.pdf.PdfWriter;
import com.itextpdf.text.pdf.BaseFont;
import com.itextpdf.text.pdf.PdfContentByte;
import com.itextpdf.text.pdf.PdfReader;
import com.itextpdf.text.pdf.PdfImportedPage;
import com.itextpdf.text.Font;
import com.itextpdf.text.Rectangle;
import com.itextpdf.text.PageSize;
import com.itextpdf.text.BadElementException;
import com.itextpdf.text.Image;
import com.itextpdf.text.BaseColor;

/**
*1ページのレイアウトを行うための基底クラスです。
*/
public abstract class PdfBuilder{

	OutputStream fOutputStream;
	org.w3c.dom.Element fTemplate;
	Document fDocument;
	PdfWriter fWriter;
	PdfContentByte fDirectContent;
	PdfContentByte fDirectContentUnder;
	String fImagePath;
	FontManager fFontManager;

	public PdfBuilder(OutputStream aOutputStream){
		try{
			fOutputStream = aOutputStream;
	        fDocument = new Document();
	        fWriter = PdfWriter.getInstance(fDocument, fOutputStream);
			fFontManager = new FontManager();
		}catch(DocumentException e){
			e.printStackTrace();
		}
	}

	public void setImagePath(String aImagePath){
		fImagePath = aImagePath;
	}

	public void setDefaultFontPath(String aDefaultFontPath){
		fFontManager.setDefaultFontPath(aDefaultFontPath);
	}

	public void setTemplateFile(File aFile){
		try{
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			org.w3c.dom.Document xmlDocument = builder.parse(aFile);
			org.w3c.dom.Element xmlElement = xmlDocument.getDocumentElement();
			setTemplate(xmlElement);
		}catch(org.xml.sax.SAXException e){
			e.printStackTrace();
			throw new IllegalStateException(e.getMessage());
		}catch(javax.xml.parsers.ParserConfigurationException e){
			e.printStackTrace();
			throw new IllegalStateException(e.getMessage());
		}catch(IOException e){
			e.printStackTrace();
			throw new IllegalStateException(e.getMessage());
		}
	}

	public void setTemplate(org.w3c.dom.Element aTemplate){
		fTemplate=aTemplate;
		Rectangle mSize = PageSize.A4;
		String mAttribute = fTemplate.getAttribute("size");
		if(!mAttribute.equals("")){
			if(mAttribute.equals("A4P")||mAttribute.equals("A4")){
				//default;
			}else if(mAttribute.equals("A4L")){
					float x = getSize("297mm");
					float y = getSize("210mm");
					mSize = new Rectangle(x,y);
			}else if(mAttribute.equals("A5")){
					float x = getSize("148.5mm");
					float y = getSize("210mm");
					mSize = new Rectangle(x,y);
			}else if(mAttribute.equals("A5L")){
					float x = getSize("210mm");
					float y = getSize("148.5mm");
					mSize = new Rectangle(x,y);
			}else{
				try{
					String[] size = mAttribute.split(" ");
					float x = getSize(size[0]);
					float y = getSize(size[1]);
					mSize = new Rectangle(x,y);
				}catch(Exception e){
					throw new IllegalArgumentException("page size not supported:" + mAttribute);
				}
			}
		}
		fDocument.setPageSize(mSize);

		float mMarginLeft=getSize("15mm");
		mAttribute  = fTemplate.getAttribute("margin-left");
		if(!mAttribute.equals("")){
			mMarginLeft=getSize(mAttribute);
		}

		float mMarginRight=getSize("15mm");
		mAttribute  = fTemplate.getAttribute("margin-right");
		if(!mAttribute.equals("")){
			mMarginRight=getSize(mAttribute);
		}

		float mMarginTop=getSize("15mm");
		mAttribute  = fTemplate.getAttribute("margin-top");
		if(!mAttribute.equals("")){
			mMarginTop=getSize(mAttribute);
		}

		float mMarginBottom=getSize("15mm");
		mAttribute  = fTemplate.getAttribute("margin-bottom");
		if(!mAttribute.equals("")){
			mMarginBottom=getSize(mAttribute);
		}

		fDocument.setMargins(mMarginLeft,mMarginRight,mMarginTop,mMarginBottom);

		NodeList listFont = fTemplate.getElementsByTagName("font");
		if(listFont.getLength()!=0){
			fFontManager.register(listFont);
		}
        fDocument.open();
		fDirectContent = fWriter.getDirectContent();
		fDirectContentUnder = fWriter.getDirectContentUnder();
	}

    public void addPage() throws IllegalStateException{
		HashMap<String,String> hm = new HashMap<String,String>();
		addPage(hm);
	}

    public abstract void addPage(HashMap<String,String> aData) throws IllegalStateException;

	public void close(){
        fDocument.close();
	}

	void processImage(boolean aBackground)throws BadElementException,DocumentException,MalformedURLException,IOException{
		NodeList images = fTemplate.getElementsByTagName("img");
		for(int i=0;i<images.getLength();i++){
			org.w3c.dom.Element item = (org.w3c.dom.Element)images.item(i);
			String mFileName = item.getAttribute("src");
			if(fImagePath!=null){
				mFileName = fImagePath + "/" + mFileName;
			}
			Image image = Image.getInstance(mFileName);
			String isBackGround = item.getAttribute("background");
			boolean mBackground = isBackGround.equals("true");
			if(mBackground == aBackground){
				float width;
				String mAttribute = item.getAttribute("width");
				if(mAttribute.equals("")){
					width=image.getWidth();
				}else{
					width=getSize(mAttribute);
				}
				float height;
				mAttribute = item.getAttribute("height");
				if(mAttribute.equals("")){
					height=image.getHeight();
				}else{
					height=getSize(mAttribute);
				}
				image.scaleAbsolute(width,height);
				float top = getSize(item.getAttribute("top"));
				float left = getSize(item.getAttribute("left"));
				image.setAbsolutePosition(left,fDocument.getPageSize().getHeight()-top-height);
				fDirectContent.addImage(image);
			}
		}
	}

	void processPDF()throws BadElementException,DocumentException,MalformedURLException,IOException{
		try{
			NodeList items = fTemplate.getElementsByTagName("pdf");
			if(items.getLength()>1){
				throw new IllegalArgumentException("only one pdf can be included.");
			}
			org.w3c.dom.Element item = (org.w3c.dom.Element)items.item(0);
			if(item==null){
				return;
			}
			String mFileName = item.getAttribute("src");
			if(fImagePath!=null){
				mFileName = fImagePath + "/" + mFileName;
			}
			String mAttribute = item.getAttribute("page");
			int mPage = 1;
			if(!mAttribute.equals("")){
				mPage=Integer.parseInt(mAttribute);
			}
			mAttribute = item.getAttribute("left");
			float mLeft = 0;
			if(!mAttribute.equals("")){
				mLeft=getSize(mAttribute);
			}
			mAttribute = item.getAttribute("top");
			float mTop = 0;
			if(!mAttribute.equals("")){
				mTop=getSize(mAttribute);
			}
			PdfReader mReader = new PdfReader(mFileName);
			PdfImportedPage mPdfPage = fWriter.getImportedPage(mReader,mPage);
			Image mImage = Image.getInstance(mPdfPage);
			mImage.setAbsolutePosition(mLeft,fDocument.getPageSize().getHeight()-mImage.getHeight()-mTop);
			fDirectContent.addImage(mImage);
		}catch(Exception e){
			if(e.getMessage().indexOf("password")!=-1){
				throw new IllegalArgumentException("PDF requires password, not supported.");
			}
			throw new IllegalArgumentException(e.getMessage());
		}
		
	}

	void processRectangle()throws BadElementException,DocumentException,MalformedURLException,IOException{
		NodeList images = fTemplate.getElementsByTagName("rect");
		for(int i=0;i<images.getLength();i++){
			fDirectContentUnder.saveState();
			org.w3c.dom.Element item = (org.w3c.dom.Element)images.item(i);
			float mLeft = getSize(item.getAttribute("left"));
			float mWidth = getSize(item.getAttribute("width"));
			float mHeight = getSize(item.getAttribute("height"));
			float mTop = fDocument.getPageSize().getHeight()-getSize(item.getAttribute("top"))-mHeight;
			
			String attribute = item.getAttribute("border-width");
			if(!attribute.equals("")){
				fDirectContentUnder.setLineWidth(getSize(attribute));
			}
			attribute = item.getAttribute("border-color");
			if(!attribute.equals("")){
				fDirectContentUnder.setColorStroke(getColor(attribute));
			}
			fDirectContentUnder.setColorFill(BaseColor.WHITE);
			attribute = item.getAttribute("color");
			if(!attribute.equals("")){
				fDirectContentUnder.setColorFill(getColor(attribute));
			}
			attribute = item.getAttribute("round");
			if(!attribute.equals("")){
				float mRound = getSize(attribute);
				fDirectContentUnder.roundRectangle(mLeft,mTop,mWidth,mHeight,mRound);
			}else{
				fDirectContentUnder.rectangle(mLeft,mTop,mWidth,mHeight);
			}
			fDirectContentUnder.fillStroke();
			fDirectContentUnder.restoreState();
		}
	}

	void processLine()throws BadElementException,DocumentException,MalformedURLException,IOException{
		NodeList images = fTemplate.getElementsByTagName("line");
		for(int i=0;i<images.getLength();i++){
			fDirectContentUnder.saveState();
			org.w3c.dom.Element item = (org.w3c.dom.Element)images.item(i);
			float mLeft = getSize(item.getAttribute("left"));
			float mTop = fDocument.getPageSize().getHeight()-getSize(item.getAttribute("top"));
			float mRight = getSize(item.getAttribute("right"));
			float mBottom = fDocument.getPageSize().getHeight()-getSize(item.getAttribute("bottom"));
			
			String attribute = item.getAttribute("width");
			if(!attribute.equals("")){
				fDirectContentUnder.setLineWidth(getSize(attribute));
			}
			attribute = item.getAttribute("color");
			if(!attribute.equals("")){
				fDirectContentUnder.setColorStroke(getColor(attribute));
			}
			fDirectContentUnder.moveTo(mLeft,mTop);
			fDirectContentUnder.lineTo(mRight,mBottom);
			fDirectContentUnder.stroke();
			fDirectContentUnder.restoreState();
		}
	}


	float getSize(String aValue){
		if(aValue.endsWith("pt")){
			String size = aValue.substring(0,aValue.length()-2);
			return Float.parseFloat(size);
		}else if(aValue.endsWith("mm")){
			String size = aValue.substring(0,aValue.length()-2);
			return 72f/25.4f*(Float.parseFloat(size));
		}else{
			throw new IllegalArgumentException("not suported value:" + aValue);
		}
	}

	BaseColor getColor(String color){
		if(color.equals("white")){
			color="#FFFFFF";
		}else if(color.equals("black")){
			color="#000000";
		}else if(color.equals("gray")){
			color="#808080";
		}else if(color.equals("red")){
			color="#FF0000";
		}else if(color.equals("green")){
			color="#00FF00";
		}else if(color.equals("blue")){
			color="#0000FF";
		}else if(color.equals("yellow")){
			color="#FFFF00";
		}
		try{
			int r =Integer.valueOf(color.substring(1,3),16);
			int g =Integer.valueOf(color.substring(3,5),16);
			int b =Integer.valueOf(color.substring(5,7),16);
			return new BaseColor(r,g,b);
		}catch(NumberFormatException e){
			throw new IllegalArgumentException("not supported color:" + color);
		}
	}

}
